/*
** Copyright (c) 2000 D. Richard Hipp
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** General Public License for more details.
** 
** You should have received a copy of the GNU General Public
** License along with this library; if not, write to the
** Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA  02111-1307, USA.
**
** Author contact information:
**   drh@hwaci.com
**   http://www.hwaci.com/drh/
**
*************************************************************************
**
** This file implements new TCL commands used by the Tcl Web Server.
*/
#include <tcl.h>
#include <ctype.h>
#include <string.h>

/*
** For convenience sake, we define the arguments to a Tcl string command.
*/
#define TCLARGS void*ClientData,Tcl_Interp*interp,int argc,char**argv
#define TCLOBJS void*ClientData,Tcl_Interp*interp,int objc,Tcl_Obj*CONST*objv

/*
** The ::tws::render command inside of an ::tws::hsubst
** appends data to the "text" DString in an instance of this structure.
** This text is then added to the results of the ::tws::hsubst.
** An instance of this structure is attached to the interpreter so that
** ::tws::render can find it.
*/
typedef struct SubstData SubstData;
struct SubstData {
  SubstData *pNext;     /* Next SubstData on a list for the interpreter */
  Tcl_DString text;     /* Data from this hsubst */
};

/*
** tclcmd: hsubst HTML
** title: Process TCL script embedded in HTML
**
** The sole argument to this command is HTML text that 
** may contain embedded TCL script.  The return value is the
** same as its input except that the embedded TCL script has been
** evaluated and evaluation result substituted in place of the embedded 
** script.
**
** In the input HTML,
** TCL script is contained within "&lt;%...%&gt;".  Whenever
** such a script is encountered, the enclosing "&lt;%" and "%&gt;" 
** markup is removed, the script is evaluated, and the result of
** the script is substituted in place of the entire "&lt;%...%&gt;"
** string.
** If the TCL script returns an error, then this command
** returns the same error.  If the TCL script terminates with a
** <b>break</b> or <b>continue</b> command, then the result is substituted
** as it normally would be, but all text after the closing "%&gt;" is
** discarded.
**
** TCL variables may be substituted by enclosing them in "&lt;$...&gt;".
** The value of the variable replaces the entire &lt;$...&gt; string.
** If the named variable does not exist, no error is generated.  A
** non-existant variable is treated like an empty string.
**
** The <b>hsubst</b> may be nested.  That is to say, the TCL script
** contained within some &lt;%...%&gt; markup of an <b>hsubst</b>
** command may contain nested calls to <b>hsubst</b>.
**
** The TWS server uses the <b>hsubst</b> command to process all dynamic
** content.  When the dispatcher determines that the content for a
** particular URL is dynamic, it reads the matching file into memory,
** passes the file's content as the argument to <b>hsubst</b> and returns
** the result to the web browser.
*/
static int HsubstCmd(TCLARGS){
  char *zIn;
  int rc = TCL_OK;
  SubstData data;

  if( argc!=2 ){
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], 
         " text\"", 0);
    return TCL_ERROR;
  }
  zIn = argv[1];
  Tcl_DStringInit(&data.text);
  data.pNext = Tcl_GetAssocData(interp, "::tws::hsubst", 0);
  Tcl_SetAssocData(interp, "::tws::hsubst", 0, &data);
  while( *zIn ){
    int i, j, c;
    char *z;
    int cEnd;

    for(i=0; (c=zIn[i])!=0; i++){
      if( c=='<' && ((c=zIn[i+1])=='%' || c=='$') ) break;
    }
    if( i>0 ) Tcl_DStringAppend(&data.text, zIn, i);
    if( zIn[i]==0 ) break;     
    j = 0;
    c = zIn[i+1];
    if( c=='$' ){
      for(j=i+2; zIn[j] && zIn[j]!='>'; j++){}
      cEnd = zIn[j];
      zIn[j] = 0;
      z = Tcl_GetVar(interp, &zIn[i+2], 0);
      zIn[j] = cEnd;
      if( z ) Tcl_DStringAppend(&data.text, z, -1);
      if( cEnd ) j++;
    }else if( c=='%' ){
      for(j=i+2; zIn[j]; j++){
        if( zIn[j]=='%' && zIn[j+1]=='>' ){
          int isComplete;
          zIn[j] = 0;
          isComplete = Tcl_CommandComplete(&zIn[i+2]);
          zIn[j] = '%';
          if( isComplete ) break;
        }
      }
      cEnd = zIn[j];
      zIn[j] = 0;
      rc = Tcl_Eval(interp, &zIn[i+2]);
      zIn[j] = cEnd;
      if( rc==TCL_ERROR ){
        Tcl_DStringFree(&data.text);
        goto done_subst;
      }
      z = Tcl_GetStringResult(interp);
      if( z && *z ) Tcl_DStringAppend(&data.text, z, -1);
      if( zIn[j] ) j += 2;
    }
    zIn += j;
    if( rc==TCL_BREAK ){
      rc = TCL_OK;
      break;
    }else if( rc!=TCL_OK ){ 
      rc = TCL_OK;
    }
  }
  Tcl_DStringResult(interp, &data.text);

done_subst:
  if( data.pNext==0 ){
    Tcl_DeleteAssocData(interp, "::tws::hsubst");
  } else {
    Tcl_SetAssocData(interp, "::tws::hsubst", 0, data.pNext);
  }
  return rc;
}

/*
** tclcmd: render TEXT
** title: Add text to the first enclosing <b>hsubst</b> command.
**
** This command is intended for use by TCL
** script contained with &lt;%...%&gt; markup of the <b>hsubst</b>
** command.  The argument to <b>render</b> is text that will be
** inserted in the output of the <b>hsubst</b>
** just before the "&lt;%" that started the TCL script.
**
** If there are multiple <b>hsubst</b> commands active when the
** <b>render</b> command is invoked, then the text is inserted into
** the inner most <b>hsubst</b>.  If <b>render</b> is invoked outside
** of any <b>hsubst</b> then it does nothing. 
**
** Consider the following example <b>hsubst</b> command:
**
** <blockquote>
** hsubst {one <% render "two"; return "three" %> four}
** </blockquote>
**
** The result of the previous command is:
**
** <blockquote>
** one twothree four
** </blockquote>
**
** The "one" and "four" at the beginning and end of the result come
** directly from the input string.  The "two" comes from the <b>render</b>
** command and the "three" is the return value of the embedded TCL
** script.
*/
static int RenderCmd(TCLOBJS){
  SubstData *pData;
  char *zOut;
  int nOut;
  if( objc!=2 ){
    Tcl_WrongNumArgs(interp, 1, objv, "text");
    return TCL_ERROR;
  }
  pData = Tcl_GetAssocData(interp, "::tws::hsubst", 0);
  if( pData ){
    zOut = Tcl_GetStringFromObj(objv[1], &nOut);
    Tcl_DStringAppend(&pData->text, zOut, nOut);
  }
  return 0;
}

/*
** Make the given string safe for HTML by converting every "<" into "&lt;",
** every ">" into "&gt;" and every "&" into "&amp;".  Return a pointer
** to a new string obtained from ckalloc().
**
** We also encode " as &quot; so that it can appear as an argument
** to markup.
*/
static char *HtmlizeString(const char *zIn){
  int c;
  int i = 0;
  int count = 0;
  char *zOut;

  while( (c = zIn[i])!=0 ){
    switch( c ){
      case '<':   count += 4;       break;
      case '>':   count += 4;       break;
      case '&':   count += 5;       break;
      case '"':   count += 6;       break;
      default:    count++;          break;
    }
    i++;
  }
  i = 0;
  zOut = ckalloc( count+1 );
  if( zOut==0 ) return 0;
  while( (c = *zIn)!=0 ){
    switch( c ){
      case '<':   
        zOut[i++] = '&';
        zOut[i++] = 'l';
        zOut[i++] = 't';
        zOut[i++] = ';';
        break;
      case '>':   
        zOut[i++] = '&';
        zOut[i++] = 'g';
        zOut[i++] = 't';
        zOut[i++] = ';';
        break;
      case '&':   
        zOut[i++] = '&';
        zOut[i++] = 'a';
        zOut[i++] = 'm';
        zOut[i++] = 'p';
        zOut[i++] = ';';
        break;
      case '"':   
        zOut[i++] = '&';
        zOut[i++] = 'q';
        zOut[i++] = 'u';
        zOut[i++] = 'o';
        zOut[i++] = 't';
        zOut[i++] = ';';
        break;
      default:
        zOut[i++] = c;
        break;
    }
    zIn++;
  }
  zOut[i] = 0;
  return zOut;
}

/*
** tclcmd: htmlize TEXT
** title: Encode characters that have special meaning to HTML
**
** This command encodes special characters in an arbitrary piece
** of input text according to the rules of HTML.  The "&lt;" character
** is converted to "&amp;lt;", the "&amp;" character is converted
** to "&amp;amp;" and the "&gt;" character is converted to "&amp;gt;".
** (The last change is not strictly necessary but is done
** for symmetry.)  In addition, the double-quote character is coded
** as "&amp;quot;" so that result can be used as an argument to markup
** if desired.
*/
static int HtmlizeCmd(TCLARGS){
  if( argc!=2 ){
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
        " STRING\"", 0);
    return TCL_ERROR;
  }
  Tcl_SetResult(interp, HtmlizeString(argv[1]), TCL_DYNAMIC);
  return TCL_OK;
}

/*
** Encode a string for HTTP.  This means converting lots of
** characters into the "%HH" where H is a hex digit.  It also
** means converting spaces to "+".
**
** This is the opposite of DeHttpizeString below.
*/
static char *HttpizeString(const char *zIn, int encodeSlash){
  int c;
  int i = 0;
  int count = 0;
  char *zOut;
  int other;
# define IsSafeChar(X)  \
     (isalnum(X) || (X)=='.' || (X)=='$' || (X)=='-' || (X)=='_' || (X)==other)

  other = encodeSlash ? 'a' : '/';
  while( (c = zIn[i])!=0 ){
    if( IsSafeChar(c) || c==' ' ){
      count++;
    }else{
      count += 3;
    }
    i++;
  }
  i = 0;
  zOut = ckalloc( count+1 );
  if( zOut==0 ) return 0;
  while( (c = *zIn)!=0 ){
    if( IsSafeChar(c) ){
      zOut[i++] = c;
    }else if( c==' ' ){
      zOut[i++] = '+';
    }else{
      zOut[i++] = '%';
      zOut[i++] = "0123456789ABCDEF"[(c>>4)&0xf];
      zOut[i++] = "0123456789ABCDEF"[c&0xf];
    }
    zIn++;
  }
  zOut[i] = 0;
  return zOut;
}

/*
** tclcmd: httpize TEXT
** title: Escape characters that have special meaning within URLs
**
** This command escapes special characters in an arbitrary piece
** of input text according to the rules of HTTP.  Spaces are converted
** to "+" and other special characters are converted into a three-letter
** code consisting of a "%" followed by two hex digits that define the
** ASCII value of the character.  The resulting string is safe to
** use as part of a URL.
*/
static int HttpizeCmd(TCLARGS){
  if( argc!=2 ){
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
        " STRING\"", 0);
    return TCL_ERROR;
  }
  Tcl_SetResult(interp, HttpizeString(argv[1],1), TCL_DYNAMIC);
  return TCL_OK;
}

/*
** tclcmd: urlize TEXT
** title: Escape characters that have special meaning within URLs
**
** This command works just like the <b>httpize</b> command except that
** it does <em>not</em> encode the "/" symbol.
*/
static int UrlizeCmd(TCLARGS){
  if( argc!=2 ){
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
        " STRING\"", 0);
    return TCL_ERROR;
  }
  Tcl_SetResult(interp, HttpizeString(argv[1],0), TCL_DYNAMIC);
  return TCL_OK;
}

/*
** Convert a single HEX digit to an integer
*/
static int AsciiToHex(int c){
  if( c>='a' && c<='f' ){
    c += 10 - 'a';
  }else if( c>='A' && c<='F' ){
    c += 10 - 'A';
  }else if( c>='0' && c<='9' ){
    c -= '0';
  }else{
    c = 0;
  }
  return c;
}

/*
** Remove the HTTP encodings from a string and construct a Tcl_Obj to
** hold the result.
*/
static Tcl_Obj *DeHttpizeString(const char *z, int n){
  int i, j;
  char *zNew = ckalloc( n+1 );
  Tcl_Obj *pObj;
  i = j = 0;

  while( i<n ){
    switch( z[i] ){
      case '%':
        if( i+2<n ){
          zNew[j] = AsciiToHex(z[i+1]) << 4;
          zNew[j] |= AsciiToHex(z[i+2]);
          i += 2;
        }
        break;
      case '+':
        zNew[j] = ' ';
        break;
      default:
        zNew[j] = z[i];
        break;
    }
    i++;
    j++;
  }
  pObj = Tcl_NewStringObj(zNew, j);
  ckfree(zNew);
  return pObj;
}

/*
** tclcmd: dehttpize TEXT
** title: Remove the special HTTP encodings on a URL or query string.
**
** This command removes the URL encodings from its input string and
** returns the result.  URL encodings are the special codes used to
** escape syntactic within a URL or query string.  For example, this
** command converts "+" into a space and "%2F" into
** a single "/" character.
*/
static int DehttpizeCmd(TCLOBJS){
  char *zOld;
  int nOld;

  if( objc!=2 ){
    Tcl_WrongNumArgs(interp, 1, objv, "STRING");
    return TCL_ERROR;
  }
  zOld = Tcl_GetStringFromObj(objv[1], &nOld);
  Tcl_SetObjResult(interp, DeHttpizeString(zOld, nOld));
  return TCL_OK;
}

/*
** x-tclcmd: decode_query TEXT
** title: Translate an HTTP query string into a TCL list
**
** The input text is the query string from a URL or from the POST
** data of an HTTP request.  The input is parsed into a sequence
** of name value pairs, the special characters are decoded, and the
** result is put into a properly formatted TCL list.
**
** For example, the query string
**
** <blockquote><pre>
** a=123.4&amp;b=%2B19&amp;msg=Hello+World%21
** </pre></blockquote>
**
** is converted into the following TCL list:
**
** <blockquote><pre>
** a   123.4   b   +19   msg   {Hello World!}
** </pre></blockquote>
**
** The TWS server uses this command internally to decode query
** strings and cookies.
*/
static int queryToList(TCLOBJS){
  char *zOld;
  int nOld;
  int i, j;
  Tcl_Obj *pResult;

  if( objc!=2 ){
    Tcl_WrongNumArgs(interp, 1, objv, "STRING");
    return TCL_ERROR;
  }
  zOld = Tcl_GetStringFromObj(objv[1], &nOld);
  pResult = Tcl_NewObj();
  i = 0;
  while( i<nOld ){
    Tcl_Obj *pElem;
    for(j=i; j<nOld && zOld[j]!='&' && zOld[j]!='=' && zOld[j]!=';'; j++){}
    pElem = DeHttpizeString(&zOld[i], j-i);
    Tcl_ListObjAppendElement(interp, pResult, pElem);
    if( j>=nOld || zOld[j]=='&' || zOld[j]==';' ){
      Tcl_ListObjAppendElement(interp, pResult, Tcl_NewObj());
    }else{
      i = j+1;
      for(j=i; j<nOld && zOld[j]!='&' && zOld[j]!=';'; j++){}
      pElem = DeHttpizeString(&zOld[i], j-i);
      Tcl_ListObjAppendElement(interp, pResult, pElem);
    }
    i = j+1;
    while( i<nOld && isspace(zOld[i]) ){ i++; }
  }
  Tcl_SetObjResult(interp, pResult);
  return TCL_OK;
}


/*
** The characters used for HTTP base64 encoding.
*/
static unsigned char zBase[] = 
  "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

/*
** tclcmd: encode64 TEXT
** title: Encode the input text using base-64
**
** This command converts the text of its input into a base-64 encoding.
** The encoding can be reversed using the <b>decode64</b> command.
*/
static int encode64Cmd(TCLOBJS){
  char *zData;
  int nData;
  char *z64;
  int i, n;

  if( objc!=2 ){
    Tcl_WrongNumArgs(interp, 1, objv, "STRING");
    return TCL_ERROR;
  }
  zData = Tcl_GetStringFromObj(objv[1], &nData);
  z64 = ckalloc( (nData*4)/3 + 6 );
  for(i=n=0; i+2<nData; i+=3){
    z64[n++] = zBase[ (zData[i]>>2) & 0x3f ];
    z64[n++] = zBase[ ((zData[i]<<4) & 0x30) | ((zData[i+1]>>4) & 0x0f) ];
    z64[n++] = zBase[ ((zData[i+1]<<2) & 0x3c) | ((zData[i+2]>>6) & 0x03) ];
    z64[n++] = zBase[ zData[i+2] & 0x3f ];
  }
  if( i+1<nData ){
    z64[n++] = zBase[ (zData[i]>>2) & 0x3f ];
    z64[n++] = zBase[ ((zData[i]<<4) & 0x30) | ((zData[i+1]>>4) & 0x0f) ];
    z64[n++] = zBase[ ((zData[i+1]<<2) & 0x3c) ];
  }else if( i<nData ){
    z64[n++] = zBase[ (zData[i]>>2) & 0x3f ];
    z64[n++] = zBase[ ((zData[i]<<4) & 0x30) ];
  }
  Tcl_SetObjResult(interp, Tcl_NewStringObj(z64, n));
  ckfree(z64);
  return TCL_OK;
}

/*
** tclcmd: decode64 TEXT
** title: Decode the input text from base-64 back to binary
**
** This command treats its input as a base-64 string and returns the
** decoded value of that string.  Characters of input that are not
** valid base-64 characters (such as spaces and newlines) are ignored.
*/
static int decode64Cmd(TCLOBJS){
  char *zData;
  char *z64;
  int n64;
  int i, j;
  int a, b, c, d;
  static int isInit = 0;
  static int trans[128];

  if( !isInit ){
    for(i=0; i<128; i++){ trans[i] = 0; }
    for(i=0; zBase[i]; i++){ trans[zBase[i] & 0x7f] = i; }
    isInit = 1;
  }
  if( objc!=2 ){
    Tcl_WrongNumArgs(interp, 1, objv, "STRING");
    return TCL_ERROR;
  }
  z64 = Tcl_GetStringFromObj(objv[1], &n64);
  while( n64>0 && z64[n64-1]=='=' ) n64--;
  zData = ckalloc( (n64*3)/4 + 4 );
  for(i=j=0; i+3<n64; i+=4){
    a = trans[z64[i] & 0x7f];
    b = trans[z64[i+1] & 0x7f];
    c = trans[z64[i+2] & 0x7f];
    d = trans[z64[i+3] & 0x7f];
    zData[j++] = ((a<<2) & 0xfc) | ((b>>4) & 0x03);
    zData[j++] = ((b<<4) & 0xf0) | ((c>>2) & 0x0f);
    zData[j++] = ((c<<6) & 0xc0) | (d & 0x3f);
  }
  if( i+2<n64 ){
    a = trans[z64[i] & 0x7f];
    b = trans[z64[i+1] & 0x7f];
    c = trans[z64[i+2] & 0x7f];
    zData[j++] = ((a<<2) & 0xfc) | ((b>>4) & 0x03);
    zData[j++] = ((b<<4) & 0xf0) | ((c>>2) & 0x0f);
  }else if( i+1<n64 ){
    a = trans[z64[i] & 0x7f];
    b = trans[z64[i+1] & 0x7f];
    zData[j++] = ((a<<2) & 0xfc) | ((b>>4) & 0x03);
  }
  Tcl_SetObjResult(interp, Tcl_NewStringObj(zData, j));
  ckfree(zData);
  return TCL_OK;
}

#ifdef TWS_SETUSER
#include <pwd.h>
#include <sys/types.h>
#include <unistd.h>
#endif

/*
** x-tclcmd: setuser USERID GROUPID
** title: Change the real and effective user
**
** On multi-user operating systems, this command attempts to convert
** the current user to the user whose name is given in the argument.
** This usually only works if the user was the superuser prior to the
** conversion.  On single-user operating systems (like Windows) this
** command is a no-op.
**
** The command is used in one place in TWS only.  TWS is usually
** started as the superuser so that it can open a listening socket
** on the standard HTTP TCP port, port 80.  After opening this socket,
** the TWS process uses this command to become an unprivileged user
** before servicing any HTTP requests.
*/
static int SetUserCmd(TCLARGS){
#ifdef TWS_SETUSER
  int uid, gid;
  if( argc!=3 ){
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
        " USERID GROUPID\"", 0);
    return TCL_ERROR;
  }
  if( Tcl_GetInt(interp, argv[1], &uid) ) return TCL_ERROR;
  if( Tcl_GetInt(interp, argv[2], &gid) ) return TCL_ERROR;
  setgid(gid);
  setuid(uid);
#endif
  return TCL_OK;
}

/*
** x-tclcmd: getuserid USERNAME
** title: Get user and group IDs given a user name
**
** On multi-user operating systems, this command returns the numeric
** user and group ids give a user name.  
*/
static int GetUserIdCmd(TCLARGS){
  char zBuf[100];
#ifdef TWS_SETUSER  
  struct passwd *pw;
  if( argc!=2 ){
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
        " USERNAME\"", 0);
    return TCL_ERROR;
  }
  pw = getpwnam(argv[1]);
  if( pw==0 ){
    Tcl_AppendResult(interp,"no such user \"", argv[1], "\"", 0);
    return TCL_ERROR;
  }
  sprintf(zBuf, "%d %d", pw->pw_uid, pw->pw_gid);
#else
  sprintf(zBuf, "1000 1000");
#endif
  Tcl_AppendResult(interp, zBuf, 0);
  return TCL_OK;
}

/*
** x-tclcmd: chroot PATH
** title: Change the root directory
**
** This command changes the root directory of the filesystem to the
** directory specified.  This command only works under UNIX when
** running as the superuser.  As any other user or on Windows, this
** command is a no-op.
**
** Return 1 on success and 0 on failure.  0 is always returned for
** Windows.
**
** The command is used in one place in TWS only.  It is used to
** put the TWS server into a chroot jail just prior to dropping
** superuser privileges.  Placing the server in a chroot jail
** is an extra security precaution.  If an attacker manages to
** break into the server (however unlikely) they will be unable
** to effect any files outside of the jail, thus the damage from
** an attack is localized and easily fixed.  As long as there are
** no SUID executables or special devices inside the jail, the
** host machine cannot be rooted.
*/
static int ChrootCmd(TCLARGS){
  char *zResult = "0";
  if( argc!=2 ){
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
      " PATH\"", 0);
    return TCL_ERROR;
  }
#ifdef TWS_CHROOT
  if( chroot(argv[1])==0 ){
    zResult = "1";
  }
#endif
  Tcl_AppendResult(interp, zResult, 0);
  return TCL_OK;
}

/*
** Initialize the new commands.
*/
int Tws_Init(Tcl_Interp *interp){
  Tcl_CreateCommand(interp, "tws::hsubst", HsubstCmd, 0, 0);
  Tcl_CreateObjCommand(interp, "tws::render", RenderCmd, 0, 0);
  Tcl_CreateCommand(interp, "tws::htmlize", HtmlizeCmd, 0, 0);
  Tcl_CreateCommand(interp, "tws::httpize", HttpizeCmd, 0, 0);
  Tcl_CreateCommand(interp, "tws::urlize", UrlizeCmd, 0, 0);
  Tcl_CreateObjCommand(interp, "tws::dehttpize", DehttpizeCmd, 0, 0);
  Tcl_CreateObjCommand(interp, "tws::encode64", encode64Cmd, 0, 0);
  Tcl_CreateObjCommand(interp, "tws::decode64", decode64Cmd, 0, 0);
  Tcl_CreateObjCommand(interp, "tws::decode_query", queryToList, 0, 0);
  Tcl_CreateCommand(interp, "tws::setuser", SetUserCmd, 0, 0);
  Tcl_CreateCommand(interp, "tws::getuserid", GetUserIdCmd, 0, 0);
  Tcl_CreateCommand(interp, "tws::chroot", ChrootCmd, 0, 0);
  Tcl_PkgProvide(interp, "tws", "1.0");
  return TCL_OK;
}
