# Copyright (c) 2000 D. Richard Hipp
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA  02111-1307, USA.
#
# Author contact information:
#   drh@hwaci.com
#   http://www.hwaci.com/drh/
#
###############################################################################
#
# A simple HTTP server useful for embedding in a larger application.
#
# The script in this file must be coupled with some new TCL commands
# implemented in C code from the file "tws.c".  All the TCL procs from
# this file, and the new TCL commands implemented by "tws.c" are contained
# in the ::tws namespace.
#
# @(#) $Id: tws.tcl,v 1.16 2001/02/17 21:16:42 drh Exp $
#

# Everything is in the ::tws namespace
#
namespace eval ::tws {

# The TWS uses the following variables from the ::tws namespace.  These
# variables should be set up before the TWS is started.
#
#     realm(A)         Information about a realm with abbreviation A.
#                         p = procedure to check credentials. {} for builtin
#                         s = require SSL if true
#                         f = full name of the realm
#
#     userpw(R U P)    Exists if P is the password for user U in realm R.
#
#     user(U)          Information about a user with user id U.
#                         n = full name of the user
#                         c = contact information
#                         w = password
#
#     extension(E)     Information about content files with extension E.
#                         h = handler. One of: dynamic static
#                         m = default mime type
#
# A typical script to set up the above control variables might look
# something like the following:
#
#   set ::tws::realm(r2) {{} 0 {Second realm}}
#   set ::tws::realm(admin) {{} 0 {TWS Administration}}
#
#   set {::tws::userpw(admin drh bogus2)} 1
#   set {::tws::userpw(r2 ggw bogus1)} 1
#   set {::tws::userpw(r2 drh bogus2)} 1
#
#   set ::tws::user(ggw) {{Ginger G. Wyrick} 704.555.1234 bogus1}
#   set ::tws::user(drh) {{D. Richard Hipp} 704.555.1234 bogus2}
#
#   set ::tws::extension(txt) {static text/plain}
#   set ::tws::extension(jpg) {static image/jpeg}
#   set ::tws::extension(exe) {static application/x-unknown}
#   set ::tws::extension(png) {static image/png}
#   set ::tws::extension(tws) {dynamic text/html}
#   set ::tws::extension(htm) {static text/html}
#   set ::tws::extension(gz) {static application/x-unknown}
#   set ::tws::extension(zip) {static application/x-unknown}
#   set ::tws::extension(html) {static text/html}
#   set ::tws::extension(gif) {static image/gif}
#   set ::tws::extension(jpeg) {static image/jpeg}
#   set ::tws::extension(bz2) {static application/x-unknown}
#

# Start a new HTTP server.  If another is running, shut it down first.
#
proc start {{p 80}} {
  halt $p
  set ::tws::sock($p) [socket -server ::tws::_accept $p]
}

# Shutdown the running HTTP server
#
proc halt {{p 80}} {
  catch {close $::tws::sock($p)}
  catch {unset ::tws::sock($p)}
}

# For a given request, the server has to determine the following information:
#
#   d:   The root of the pathname for a file or variable used to satisfy
#        the request.
#
#   a:   The realm to use for authentication.  {} means do not authenticate.
#
#   n:   The namespace in which hsubst commands run.
#
# The parameters above are collected into a list labeled $x.
#
# The parameters above are looked up from a table with a two-part key.
# The key values are:
#
#   s:   The name of the server to which the request was submitted.
#        The name must match exactly, except {} will match anything.
#
#   p:   A prefix for the path of the URI.  If multiple prefixes match
#        the URI we choose the longest prefix.
#
# The following routine registers information used to translate the s-p
# key of an HTTP request into the d-a-n result.  A dispatcher table entry
# can be deleted by setting the 3rd argument to this routine to {}.
#
# Usage Examples:
#
# The following command causes HTTP requests of the form 
# "http://localhost/config/..." to be handled by documents rooted
# at /-builtin-/config/...  The realm is {} so no authentication is
# required.  All dynamic content is processed in the ::twscfg namespace.
# Notice that the host field of the HTTP request must be "localhost"
# for this pattern to match.
#
#     ::tws::dispatcher localhost /config {/-builtin-/config {} ::twscfg}
#
# The next examples handles all requests of the form
# "http://*/acct/...".  In the previous example, the host field of
# the HTTP request had to be "localhost" but in this example the host
# field can be anything.  The request is handled using a document tree
# rooted at /home/drh/biz/acct/www.  All requested must carry valid
# authenication creditials for the realm "acct".  Dynamic content
# is evaluated in the global namespace.
#
#     ::tws::dispatcher {} /acct {/usr/home/drh/biz/acct/www acct {}}
#
# The final example deletes the dispatcher table entry that was
# created by the previous example:
#
#     ::tws::dispatcher {} /acct {}
#
proc dispatcher {s p x} {
  if {[regexp { } $s$p]} {
    error "no spaces allows an server or path"
  }
  set key [list $s $p]
  if {"$x"==""} {
    catch {unset ::tws::dispatcher($key)}
  } else {
    set ::tws::dispatcher($key) [list $x]
  }
  catch {unset ::tws::switch}
}

# Return a list of all registered dispatchers.  This routine can be used
# to query the configuration of a server in order to save its state.
#
proc dispatcher_list {} {
  set res {}
  foreach term [::tws::_dispatcher_orderedlist] {
    lappend res [lrange $term 1 end]
  }
  return $res
}

# Given a server name (s) and the complete path of the URL (u),
# locate the dispatch information for this URL.  The
# data returned is:
#
#     p = The path prefix that matched against $u
#     d = The root name of the content
#     a = The authentication realm
#     n = The namespace in which to run hsubst commands
#
# This is the key routine used to determine how incoming HTTP requests
# are handled by the TWS server.
#
proc _dispatcher_find {s u} {
  if {![info exists ::tws::switch]} ::tws::_build_switch
  set key "$s $u"
  switch -regexp -- $key $::tws::switch
  return [concat [list $p] $x]
}

# Construct a list of all dispatcher templates in matching order.  The
# most specific patterns appear first on the list.
#
# Each element of the returned list is of the form {sk s p x} where
#
#     sk =  A sort key which can be ignored
#     s  =  The server name
#     p  =  The path prefix
#     x  =  The data payload
# 
proc _dispatcher_orderedlist {} {
  upvar 0 ::tws::dispatcher V
  set list {}
  foreach k [array names V] {
    foreach {s p} $k {}
    set sk [string length $p]
    if {[string length $s]>0} {incr sk 10000}
    set sk [expr {1000000-$sk}]
    set term [concat [list $sk] $k $V($k)]
    lappend list $term
  }
  return [lsort -dictionary $list]
}

# Construct the switch statements that are used by _dispatcher_find
# to locate a dispatcher based on server name and path prefix.
#
# This routine is called the first time _dispatcher_find is invoked
# after a new dispatcher has been added using the dispatcher proc.
# Once the switch statement has been created, it is used for all
# subsequent HTTP requests until the dispatcher table changes again.
#
proc _build_switch {} {
  set ::tws::switch {}
  set dflt_x {}
  foreach term [::tws::_dispatcher_orderedlist] {
    foreach {i s p x} $term {}
    if {"$s$p"==""} {
      set dflt_x $x
      continue
    }
    regsub -all {[][*.?+()]} $s {\\&} s
    regsub -all {[][*.?+()]} $p {\\&} p
    if {$s==""} {
      set s {[^ ]*}
    } elseif {![regexp {:[0-9]+$} $s]} {
      append s {(:[0-9]+)?}
    }
    set p [string trimright $p /]
    set re [format {^%s %s(/.*)?$} $s $p]
    lappend ::tws::switch $re "set x [list $x]; set p [list $p]"
  }
  lappend ::tws::switch default "set x [list $dflt_x]; set p {}"
}

# Accept a new connection.
#
# This routine is called when an HTTP client makes its initial connection
# to the server.  All data for this connection will be recorded in a
# variable that has the same name as the socket.
#
proc _accept {chan addr host} {
  fconfigure $chan -blocking 0 -translation {auto binary}
  fileevent $chan readable [list ::tws::_read_header $chan]
  upvar 0 ::tws::$chan D
  set D(method) {}
  set D(prior) {}
  set D(count) 1
}

# Read HTTP request header data from a connection.  Store the request
# parameters in an array variable named ::tws::$sock.  Invoke
# _process_header after a complete request header has been read,
# but before the POST data is read.  (_process_header will take
# care of reading the POST data, if any.)
#
# This routine does not block.  It will execute multiple times to
# read bits of a request as the request arrives.  State information
# is stored in an array named ::tws::$sock which is aliased to
# "D" by the first statement of this routine.  Some of the key
# fields in this array include:
#
#    D(method)       One of GET, HEAD, POST, etc.
#
#    D(path)         The path of the URI
#
#    D(content)      The part of the request URI that follows the "?".
#                    The query string of the URI.  Does not exist if
#                    there is no query string on the URI.
#
#    D(protocol)     HTTP/1.0, HTTP/1.1, etc.
#
#    D(prior)        Content of previous header line.  The next header
#                    is appended if it begins with whitespace.  A line
#                    of the header is not processed until we are sure
#                    there are no more continuation lines.
#
#    D(FIELD)        Any "FIELD:" in the header has its value placed
#                    in an entry of D() whose index is the field name.
#                    The field name is converted to all upper case letters
#                    before being used as the key.  That way, no field
#                    field name will ever collide with an internally
#                    generated value like D(prior).
#
proc _read_header {sock} {
  upvar 0 ::tws::$sock D
  if {[eof $sock]} { 
    close $sock
    catch {unset D}
    return
  }
  if {[catch {gets $sock line}]} {
    close $sock
    catch {unset D}
    return
  }
  if {$D(method)==""} {
    if {$line==""} return
    set D(method) [lindex $line 0]
    set uri [lindex $line 1]
    set D(host) {}
    set D(content) {}
    if {[regexp {^([Hh][Tt][Tt][Pp]:)?(//[^/]*)?(.*)$} $uri all p h u]} {
      set uri $u
      set D(host) [string trimleft $h /]
    }
    if {[regexp {^([^?]*)(\?.*)$} $uri all u e]} {
      set uri $u
      set D(content) [string range $e 1 end]
    }
    set D(path) $uri
    set D(protocol) [lindex $line 2]
  }
  set c1 [string index $line 0]
  if {$c1==" " || $c1=="\t"} {
    append D(prior) $line
  } elseif {$D(prior)!=""} {
    if {[regexp {^ *([-a-zA-Z0-9]+): *(.*)$} $D(prior) all name value]} {
      set name [string toupper $name]
      if {[info exists D($name)]} {
        append D($name) ", [string trim $value]"
      } else {
        set D($name) [string trim $value]
      }
    }
    set D(prior) {}
  }
  if {$line==""} {
    fileevent $sock readable {}
    catch {unset D(prior)}
    ::tws::_process_header $sock
  } else {
    set D(prior) $line
  }
}

# x-tclcmd: chroot_fixup VARIABLE
# title: Remove the chroot prefix from the given variable
#
# If the TWS server is not running in a chroot jail, this routine
# is a no-op.  If TWS is in a chroot jail and VARIABLE contains
# a pathname that begins with the $::tws::chrootjail prefix, then
# the prefix is removed from the variable.
#
if {[info exists chrootjail] && $chrootjail!=""} {
   regsub -all {[][()*+.$\\]} $chrootjail {\\&} reg
   proc chroot_fixup {vname} [subst {
      upvar 1 \$vname v
      regsub {^$reg} \$v {} v
   }]
   unset reg
} else {
   proc chroot_fixup {vname} {}
}


# This routine runs after a the HTTP request header has been received.
# All of the data for the request is contained in the ::tws::$sock 
# array variable, which is aliased to the D() array inside the procedure.
# Every optional field of the HTTP header is stored as an entry in the
# D() array.  If the name of the field is ABC, then the value is stored
# in D(ABC).  The field names are converted to all caps before begin
# used for keys into the D() array.
#
# For example, if the header includes a "Host: www.hwaci.com" line, then
# the D(HOST) variable will be set to "www.hwaci.com".
#
# Additional information is stored in D() variable with lower-case names.
# There is never any confusion between this meta information and the header
# fields because the meta information always has lower case keys and the
# header fields always have upper case keys.  The available meta information
# is as follows:
#
#    D(host)           The name of this server as it appears in the
#                      URI on the first line of the HTTP request.  Normally
#                      the host name is omitted from the URI so the value
#                      of this variable is the empty string.
#
#    D(method)         "GET" or "POST" or "HEADER" or whatever the first
#                      word of the first line of the header happens to be.
#
#    D(content)        If there is a query string appended to the URI on the
#                      first line of the header, that query string is placed
#                      in this variable.
#
#    D(path)           The URI string from the first line of the header, but
#                      with the query string removed.
#
#    D(protocol)       "HTTP/1.0" or "HTTP/1.1" or whatever is the third
#                      token on the first line of the header.
# 
# If there is data following the header of the request, that data remains
# unread at the time this procedure is called.
#
# This procedure figures out how to process the request.  Here are the steps:
#
# 1.  Given the server name (in $D(host) or $D(HOST)) and the URI (in $D(path)),
#     find a document directory (stored in D(docroot), an authentication 
#     realm (stored in D(realm)), and namespace (in D(namespace) for use on
#     this request.
#
# 2.  If there is a non-null authentication realm, check the authentication
#     credentials.  Fail immediately if the credentials are not correct.
#
# 3.  Optionally change the request path to all lower-case letters.
#
# 4.  Compute the name of the file or TCL variable that will be used to
#     satisfy the request.
#
#     A.  Set D(prefix) to the part of the original URI path that matches
#         the prefix of the dispatcher table.  And set D(suffix) to the
#         part of the original URI that extends beyond the prefix.  Any
#         filename extension (ex: ".html", ".jpg", etc) that may have been
#         present on the URI is discarded.
#
#     B.  Set the document name (in D(document)) to the document directory
#         followed by the suffix.  (Note: if the document directory begins
#         with "::" then we are really dealing with the name of a TCL 
#         variable.  But we will still call it a file for simplicity of
#         nomiclature.)
#
#     C.  If the document name ends in "/", append the name "index".
#
#     D.  Check for files with the document name having any file extension.
#         If multiple files are found, chose the first in lexical order.
#
#     E.  If no files are found in step 4D, then replace the suffix in
#         the document name with "notfound" and repeat the search.
#
#     F.  If no files are found in step 4E, set the document name to
#         "::tws::_notfound"
#
# 5.  Using the extension on the document file name found in step 4 to
#     look up an appropriate document handler (static or dynamic) and 
#     a default MIME-type for the document.
#
# After the steps above are completed, request data is read (if any) and
# then the appropriate document handler is invoked.
#
proc _process_header sock {
  upvar 0 ::tws::$sock D

  # Step 1.
  if {![info exists D(HOST)]} {set D(HOST) $D(host)}
  set D(HOST) [string tolower $D(HOST)]
  if {[regexp {^localhost} $D(HOST)]} {
    set addr [lindex [fconfigure $sock -peername] 0]
    set myaddr [lindex [fconfigure $sock -sockname] 0]
    if {$addr!="127.0.0.1" && $addr!=$myaddr} {
      set D(HOST) {**impostor**}
    }
  }
  set dispatch [::tws::_dispatcher_find $D(HOST) $D(path)]
  set D(prefix) [lindex $dispatch 0]
  set D(docroot) [lindex $dispatch 1]
  chroot_fixup D(docroot)
  set D(realm) [lindex $dispatch 2]
  set D(namespace) [lindex $dispatch 3]

  # Step 2.
  if {$D(realm)!="" && ![::tws::_authenticate $sock]} {
    return
  }

  # Steps 4A, 4B, and 4C
  set D(suffix) [string range $D(path) [string length $D(prefix)] end]
  set D(suffix) [file rootname $D(suffix)]
  set D(document) $D(docroot)/$D(suffix)
  regsub -all {//+} $D(document) / D(document)
  if {[regexp {/$} $D(document)]} {
    append D(document) index
  }

  # Step 4D
  set isfile [expr {[string range $D(document) 0 1]!="::"}]
  set pattern $D(document)*
  if {$isfile} {
    set flist [concat [zvfs::list $pattern] [glob -nocomplain $pattern]]
  } else {
    set flist [info vars $pattern]
  }

  # Step 4E
  if {[llength $flist]==0} {
    set D(document) $D(docroot)/notfound
    set pattern $D(document)*
    if {$isfile} {
      set pattern 
      set flist [concat [zvfs::list $pattern] [glob -nocomplain $pattern]]
    } else {
      set flist [info vars $pattern]
    }
  }

  # Step 4F
  if {[llength $flist]==0} {
    set flist ::tws::_notfound
    set isfile 0
  }
  set D(document) [lindex [lsort $flist] 0]

  # Step 5
  set ext [string trimleft [file extension $D(document)] .]
  if {[info exists ::tws::extension($ext)]} {
    set D(handler) [lindex $::tws::extension($ext) 0]
    set D(default-mime-type) [lindex $::tws::extension($ext) 1]
  } elseif {$flist=="::tws::_notfound"} {
    set D(handler) dynamic
    set D(default-mime-type) text/html
  } else {
    set D(handler) static
    set D(default-mime-type) text/plain
  }
  if {$isfile} {set D(handler) f$D(handler)} {set D(handler) v$D(handler)}

  # Read the request data and invoke the appropriate handler
  if {[info exists D(CONTENT-LENGTH)]} {
    set D(togo) $D(CONTENT-LENGTH)
    set D(content) {}
    fileevent $sock readable "::tws::_read_data $sock"
    return
  } else {
    ::tws::_process_data $sock
  }
}

# Fill this variable with a list of Haiku error message for the
# "Not Found" reply.  One of the haikus will be selected at random.
#
variable _notfound_haiku {
  {
    You step in the stream,<br>
    but the water has moved on.<br>
    The page is not here.
  }
  {
    The page is not here.<br>
    Your expensive computer<br>
    has become useless.
  }
  {
    The page that you seek<br>
    is not found on my disk drive.<br>
    You must look elsewhere.
  }
}

# This HTML is generated if nothing else is found.
#
variable _notfound {
<html>
<head><title>Not Found</title></head>
<body bgcolor="#ffb6c1">
<h1>Not Found</h1>
<blockquote><i>
<%
  set n [llength ${::tws::_notfound_haiku}]
  set i [expr {int(rand()*$n)}]
  lindex [set ::tws::_notfound_haiku] $i
%>
</i></blockquote>
</body>
</html>
}

# This routine is call to check authentication credentials.
#
# If the authentication checks out, return TRUE. If it does not,
# generate a reply that rejects the request.  The POST data has not
# been read, so be sure to close the connection on a rejected reply.
# Return FALSE on a rejection.
#
# The credentials are checked in a particular "realm".  For each realm A
# there is a variable ::tws::realm(A) containing three terms:
#
#        p = The procedure to verify user credentials in that realm.  {}
#            is a default procedure.
#        s = TRUE if SSL is required in this realm
#        f = The full descriptive name of the realm.  "A" is just an
#            abbreviation.
#
# The default authentication is to check for the existance of a variable
# named ::tws::userpw([list $realm $user $password]).
#
proc _authenticate {sock} {
  upvar 0 ::tws::$sock D
  set realm $D(realm)
  if {![info exists ::tws::realm($realm)]} {
    ::tws::_do_notauthorized $sock $realm
    return 0
  }
  set aproc [lindex $::tws::realm($realm) 0]
  set sslreq [lindex $::tws::realm($realm) 1]
  set realmname [lindex $::tws::realm($realm) 2]
  if {![info exists D(AUTHORIZATION)]} {
    ::tws::_do_notauthorized $sock $realmname
    return 0
  }
  set userpw [split [::tws::decode64 [lindex $D(AUTHORIZATION) 1]] :]
  set D(user) [lindex $userpw 0]
  set pw [lindex $userpw 1]
  if {$aproc==""} {
    set key [list $realm $D(user) $pw]
    set res [info exists ::tws::userpw($key)]
  } elseif {[catch {$aproc $realm $D(user) $pw} res]} {
    set res 0
  }
  if {!$res} {
    ::tws::_do_notauthorized $sock $realmname
  }
  return $res
}

# This routine notifies the client that a request is denied
# because the request lacks proper authentication credentials.
#
proc _do_notauthorized {sock realm}  {
  upvar 0 ::tws::$sock D
  set D(reply-content) [hsubst {<html>
<body bgcolor="white" text="black">
<h1>Unauthorized</h1>
<p>The request lacks valid authentication credentials.</p>
</body></html>
}]
  set D(CONTENT-TYPE) text/html
  set D(WWW-AUTHENTICATE) "Basic realm=\"[httpize $realm]\""
  set D(status-code) {401 Unauthorized}
  ::tws::_reply $sock
}

# Read the POST data.  After it has all be read, invoke _process_data
# to complete the processing of this request.
#
# The post data is stored in D(content).  D(togo) holds the
# number of bytes of POST data yet to be read.
#
# This routine is non-blocking.  It is called when there is more
# data available to be read.  It might be called multiple times before
# all of the POST data has been loaded.
#
proc _read_data sock {
  upvar 0 ::tws::$sock D
  if {[eof $sock]} { 
    close $sock
    catch {unset D}
    return
  }
  append D(content) [read $sock $D(togo)]
  set sofar [string length $D(content)]
  set D(togo) [expr {$D(CONTENT-LENGTH)-$sofar}]
  if {$D(togo)==0} {
    fileevent $sock readable {}
    ::tws::_process_data $sock
    return
  }
}

# This routine runs after the POST data has been read.  The header
# was processed before the POST data was read.  All that is left to
# do is invoke the appropriate handler.
#
# If a handler throws an exception, check the error message.  The
# "redirect" command works by throwing a special exception.  If the
# error message indicates a redirect, then do the redirect.  Otherwise,
# return an error message page to the client.
#
proc _process_data sock {
  upvar 0 ::tws::$sock D
  if {![info exists D(content)]} {set D(content) {}}
  set D(query) [::tws::decode_query $D(content)]
  if {[info exists D(COOKIE)]} {
     set D(cookie) [::tws::decode_query $D(COOKIE)]
  } else {
     set D(cookie) {}
  }
  catch {set D(content-type) $D(CONTENT-TYPE); unset D(CONTENT-TYPE)}
  if {[catch {eval ::tws::_do_$D(handler) $sock} msg]} {
    if {[regexp {^-tws-reloc-(.*)$} $msg all path]} {
      if {![regexp -nocase {^(http:/)?/} $path]} {
        set path [file dirname $D(path)]/$path
        regsub -all {/\./} $path / path
        while {[regsub -all {[^/]*[^/.]/\.\./} $path {} path]} {}
        regsub {^(\./)+} $path {} path
        regsub -all {//+} $path {/} path
      }
      set D(reply-content) {<html><p>Moved Temporarily</p></html>}
      set D(LOCATION) $path
      set D(status-code) "302 Moved Temporarily"
      ::tws::_reply $sock
    } else {
      ::tws::_do_error $sock
    }
  }
}

# This routine sends an error message reply when a handler barfs
#
proc _do_error sock {
  upvar 0 ::tws::$sock D
  set D(reply-content) [hsubst {<html>
<body bgcolor="white" text="black">
<h1>Internal Server Error</h1>
<pre>
<% htmlize $::errorInfo %>
</pre>
</body></html>
}]
  set D(CONTENT-TYPE) text/html
  set D(status-code) {500 Internal Server Error}
  catch {unset D(send-file)}
  ::tws::_reply $sock
}

# Copy a TCL variable as the reply
#
proc _do_vstatic sock {
  upvar 0 ::tws::$sock D
  set D(reply-content) [set $D(document)]
  ::tws::_reply $sock
}

# Do an "hsubst" on a TCL variable to form the reply
#
proc _do_vdynamic sock {
  upvar 0 ::tws::$sock D
  if {$D(namespace)==""} {
    array set Q [concat $D(cookie) $D(query)]
    set D(reply-content) [hsubst [set $D(document)]]
  } else {
    set pname ${D(namespace)}::__tws_hsubst
    if {[info procs $pname]==""} {
      proc $pname {svn dvn} {
        upvar 0 $svn D
        array set Q [concat $D(cookie) $D(query)]
        return [::tws::hsubst [set $dvn]]
      }
    }
    set D(reply-content) [$pname ::tws::$sock $D(document)]
  }
  ::tws::_reply $sock
}

# Copy a disk files as the reply
#
proc _do_fstatic sock {
  upvar 0 ::tws::$sock D
  set D(send-file) $D(document)
  ::tws::_reply $sock
}

# Do a "hsubst" on a disk file to form the reply
#
proc _do_fdynamic sock {
  upvar 0 ::tws::$sock D
  set fn $D(document)
  set D(pwd) [file dirname $fn]
  set fd [open $fn r]
  set D(reply-content) [read $fd [file size $fn]]
  set __private_oldpwd [pwd]
  catch {cd $D(pwd)}
  close $fd
  unset fd
  unset fn
  if {$D(namespace)==""} {
    array set Q [concat $D(cookie) $D(query)]
    set D(reply-content) [hsubst $D(reply-content)]
  } else {
    set pname ${D(namespace)}::__tws_hsubst
    if {[info procs $pname]==""} {
      proc $pname {svn dvn} {
        upvar 0 $svn D
        array set Q [concat $D(cookie) $D(query)]
        return [::tws::hsubst [set $dvn]]
      }
    }
    set D(reply-content) \
       [$pname ::tws::$sock ::tws::${sock}(reply-content)]
  }
  catch {cd $__private_oldpwd}
  ::tws::_reply $sock
}

# When the processing of an HTTP request is complete, the handler
# should invoke this routine to send the reply.  All the information
# needed to construct the reply is contained in the ::tws::$sock
# array variable.  The following fields are used:
#
#    D(reply-content)       The content of the reply message
#    D(status-code)         Status code for the reply message
#    D(CONNECTION)          "keep-alive" or "close"
#    D(CONTENT-TYPE)        Mime type of the reply
#    D(CONTENT-ENCODING)
#    D(CONTENT-LENGTH)      Computed automatically from reply-content
#    D(DATE)
#    D(EXPIRES)
#    D(LOCATION)            Alternate URL if status-code is 301 or 302
#    D(PRAGMA)
#    D(SERVER)              A default is supplied if omitted
#    D(SET-COOKIE)
#    D(WWW-AUTHENTICATE)
# 
proc _reply sock {
  upvar 0 ::tws::$sock D

  # Prepare variables
  #
  if {![info exists D(status-code)]} {set D(status-code) {200 OK}}
  if {![info exists D(CONTENT-TYPE)]} {
    set D(CONTENT-TYPE) $D(default-mime-type)
  }
  if {[info exists D(debug)] && $D(debug)!=""
      && [regexp {^text/} $D(CONTENT-TYPE)]} {
    set D(reply-content) "[string trim $D(debug)]\n<hr>\n$D(reply-content)"
    set D(status-code) {200 OK}
  }
  if {[info exists D(send-file)]} {
    if {[catch {open $D(send-file) r} D(send-file-fd)]} {
      set ::errorInfo $D(send-file-fd)
      unset D(send-file)
      ::tws::_do_error $sock
      return
    }
    catch {unset D(reply-content)}
    fconfigure $D(send-file-fd) -translation binary
    set D(CONTENT-LENGTH) [file size $D(send-file)]
  } elseif {[info exists D(reply-content)]} {
    set D(CONTENT-LENGTH) [string length $D(reply-content)]
  } else {
    set D(CONTENT-LENGTH) 0
  }
  if {![info exists D(SERVER)]} {
    set D(SERVER) {TclServer version 0.01}
  }

  # Send the reply header
  #
  puts -nonewline $sock "HTTP/1.0 $D(status-code)\r\n"
  foreach f {
    CONTENT-ENCODING
    CONTENT-LENGTH
    CONTENT-TYPE
    DATE
    EXPIRES
    LOCATION
    PRAGMA
    SERVER
    WWW-AUTHENTICATE
  } {
    if {[info exists D($f)]} {
      puts -nonewline $sock "$f: $D($f)\r\n"
    }
  }
  if {[info exists D(set-cookie)]} {
    foreach c $D(set-cookie) {
      puts -nonewline $sock "SET-COOKIE: $c\r\n"
    }
  }
  puts -nonewline $sock "\r\n"
  if {[info exists D(send-file)]} {
    fcopy $D(send-file-fd) $sock -command "::tws::_reply_finalize $sock"
  } elseif {[info exists D(reply-content)]} {
    puts -nonewline $sock $D(reply-content)
    flush $sock
    ::tws::_reply_finalize $sock
  }
}

# This routine is a callback that _reply invokes after it finishes
# sending all of the reply data.
#
proc _reply_finalize {sock args} {
  upvar 0 ::tws::$sock D

  catch {close $D(send-file)}

  # Figure out if we should close the
  # connection or wait on another request.
  #
  if {[catch {scan $D(protocol) {%*[^/]/%f} vers}]} {set vers 0.9}
  if {$vers>=1.1} {
    if {[info exists D(connection)] && $D(connection)=="close"} {
      close $sock
      unset D
      return
    }
  } elseif {$vers==1.0} {
    if {![info exists D(connection)] || $D(connection)=="close"} {
      close $sock
      unset D
      return
    }
  } else {
    close $sock
    unset D
    return
  }
  fileevent $sock readable [list ::tws::_read_header $sock]
  set cnt $D(count)
  unset D
  set D(count) [expr {$cnt+1}]
  set D(method) {}
  set D(prior) {}
}

###############################################################################
# Helper methods for the HTTP server.
#
# These are non-essential "nice-to-have" functions.
#

# tclcmd: require FILENAME
# title: Source the named TCL script if it has not been sourced already
#
# This command is available for use by embedded TCL scripts of dynamic
# content files.  The purpose of this command is to make sure that a
# file of TCL code has been sourced and that all the procedures and
# variables defined by that file are available to the embedded TCL script.
#
# The name of the TCL file to be sourced is given as an argument.  The
# name may be either an absolute or a relative pathname.  If it is a
# relative pathname, then the name is relative to the directory that
# contains the original dynamic content file.
#
# The named file is sourced only once, the first time that the <b>require</b>
# command is executed.  On subsequent executions, the <b>require</b>
# command is essentially a no-op.
#
proc require {filename} {
  upvar 1 D D
  if {[file pathtype $filename]!="absolute"} {
     set filename $D(pwd)/$filename
  }
  regsub -all {//+} $filename / filename
  regsub -all {/\./} $filename / filename
  while {[regsub -all {[^/]*[^/.]/\.\./} $filename {} filename]} {}
  set ns $D(namespace)
  if {![info exists ::tws::source($ns,$filename)] ||
       [file mtime $filename]>$::tws::source($ns,$filename)} {
    if {$ns==""} {
      uplevel 1 "::source [list $filename]"
    } else {
      namespace eval $ns [list uplevel 1 [list ::source $filename]]
    }
    set ::tws::source($ns,$filename) [file mtime $filename]
  }
}

# tclcmd: redirect URL
# title: Cause the requesting web browser to jump to a different URL
#
# This command, when executed by an embedded TCL script of a dynamic
# content file, causes the generation of dynamic content to stop
# immediately.  The web server then sends the requesting web browser
# a message telling it to try again at a different URL specified
# in the argument.
#
# The URL argument may be either absolute or relative.  If it is
# relative, the URL of the file that executed the redirect command
# is take as the base.
#
proc redirect path {
  error "-tws-reloc-$path"
}

# Return HTML text that shows the value of all variables in the 
# calling function.
#
# This routine is useful for debugging.
#
proc list_variables {} {
  set x {}
  foreach v [uplevel 1 {lsort -dictionary [info vars]}] {
    set V ${v}_
    upvar 1 $v $V
    if {[catch {set $V} val]} {
      foreach e [lsort -dictionary [array names $V]] {
        append x [htmlize "${v}($e) = [set ${V}($e)]"]<br>\n
      }
    } else {
      append x [htmlize "$v = $val"]<br>\n
    }
  }
  return $x
}

# tclcmd: select OPTIONS...
# title: Generate <SELECT> HTML code
#
# This convenience routine
# return HTML code to implement a <SELECT>...</SELECT>.  The features of
# the select are controlled by switches as follows:
#
# <dl><dt> -name VAR </dt><dd><p>
#
#       The name of the output variable for the select statement.
#
# </dd><dt> -initial NAME </dt><dd><p>
#
#       NAME is a query parameter that determines which of the
#       options is initially selected.
#
# </dd><dt> -size N </dt><dd><p>
#
#       Number of lines in the selector.
#
# </dd><dt> -multiple BOOL </dt><dd><p>
#
#     If BOOL is non-zero, allow multiple selections.
#
# </dd><dt> -choices LIST </dt><dd><p>
#
#     LIST is a list of possible options.  Each element of
#     the list is a sublist of two elements.  The first subitem
#     is the value that VAR is set to if the option is chosen.
#     The second subitem is the text displayed to the user.
#     If there is only one subitem, it serves both roles
#
# </dd></dl>
#
proc select {args} {
  set v(-name) {}
  set v(-initial) {}
  set v(-choices) {}
  set v(-size) 1
  set v(-multiple) 0
  array set v $args
  if {$v(-multiple)} {
    set m " multiple"
    if {$v(-size)<3} {set v(-size) 3}
  } else {
    set m ""
  }
  set x "<select name=\"$v(-name)\" size=\"$v(-size)\"$m>\n"
  foreach c $v(-choices) {
    set val [lindex $c 0]
    set text [lindex $c 1]
    if {$text==""} {set text $val}
    if {[lsearch $v(-initial) $val]>=0} {
      set s { selected}
    } else {
      set s {}
    }
    append x "<option value=\"$val\"$s>$text</option>\n"
  }
  append x "</select>"
  return $x
}

# tclcmd: setcookie OPTIONS...
# title: Set a cookie
#
# Use this routine to set a cookie on the requesting web browser.
# The name and value of the cookie are required arguments.
# Other options are available.
#
#
# <dl><dt>  -name NAME  </dt><dd><p>
#
#     The name of the cookie.  This is a required field.
#
# </dd><dt> -value VALUE </dt><dd><p>
#
#     The value of the cookie.  The cookie value must be
#     a well-formed cookie string.  That is to say, the value
#     must not contain characters that have special meaning
#     in the HTTP protocol.  Use the encode64 command
#     to encode the string if there is any doubt.
#
# </dd><dt> -path PATH </dt><dd><p>
#
#     The path for which the cookie is valid.  If omitted,
#     the path defaults to "/".
#
# </dd><dt> -lifetime SECONDS </dt><dd><p>
#
#     The number of seconds until the cookie expires.  The
#     default is for the cookie to never expire.
#
# </dd><dt> -comment TEXT </dt><dd><p>
#
#     A comment describing what the cookie does.  Some web browser
#     might display this comment to the user and ask them if they
#     want to accept the cookie.
#
# </dd></dl>
#
proc setcookie {args} {
  set v(-name) {}
  set v(-value) {}
  set v(-path) /
  array set v $args
  if {$v(-name)==""} {
    error "No cookie name given"
  }
  set cookie "$v(-name)=[httpize $v(-value)]; PATH=$v(-path)"
  if {[info exists v(-lifetime)]} {
    append cookie "; MAX-AGE=$v(-lifetime)"
  }
  if {[info exists v(-comment)]} {
    append cookie "; COMMENT=[httpize $v(-comment)]"
  }
  upvar 1 D D
  lappend D(set-cookie) $cookie
}

# tclcmd: include FILENAME
# title: Include the text of another file in the middle of an HTML document.
#
# This routine reads all the text form a file, does an <b>hsubst</b>
# on that text, then returns the results.  This is useful for
# including the body of one file inside of another.
#
proc include {filename} {
  if {[file pathtype $filename]!="absolute"} {
     upvar 1 D D
     set filename $D(pwd)/$filename
  }
  set fd [open $filename r]
  set body [read $fd [file size $filename]]
  close $fd
  set pwd $D(pwd)
  set D(pwd) [file dirname $filename]
  if {$D(namespace)==""} {
    set result [uplevel 1 [list ::tws::hsubst $body]]
  } else {
    set result [namespace eval $D(namespace) \
         [list uplevel 1 [list ::tws::hsubst $body]]]
  }
  set D(pwd) $pwd
  return $result
}

# This routine does the work of an %if, %ifdef, or %ifndef statement.
# See the documention below for additional information.
#
proc _if_processor alist {
  upvar 2 Q Q
  for {set i 0} {1} {incr i} {
    set skip 0
    switch -- [lindex $alist $i] {
      %elseif {
        incr i
        set expr [lindex $alist $i]
        incr i
        set skip [expr {![uplevel 2 [list expr $expr]]}]
      }
      %elseifdef {
        incr i
        set vlist [lindex $alist $i]
        incr i
        set skip 0
        foreach v $vlist {
          if {![info exists Q($v)]} {set skip 1; break}
        }
      }
      %elseifndef {
        incr i
        set vlist [lindex $alist $i]
        incr i
        set skip 0
        foreach v $vlist {
          if {[info exists Q($v)]} {set skip 1; break}
        }
      }
      %else {
        incr i
      }
    }
    if {!$skip} {
      return [uplevel 2 [list ::tws::hsubst [lindex $alist $i]]]
    }
  }
}

# tclcmd: %if CONDITION HTML ...
# title: Conditionally include HTML text.
#
# There are really three commands with similar operation:
# <b>%if</b>, <b>%ifdef</b> and <b>%ifndef</b>.
# Each of these commands
# is designed to conditionally include some HTML text within
# dynamically generated content.  The operation of <b>%if</b> is similar to
# the standard <b>if</b> command of TCL except that the body is
# returned as HTML (after being processed by <b>hsubst</b>) instead
# of being executed as a script.
#
# The easiest way to describe the operation of these statements is by
# example.  You can put the following kinds of text in an dynamic
# content file:
#
# <blockquote><pre>
#     &lt;%%if {$value&gt;10} {
#         &lt;p&gt;The value &lt;$value&gt; is greater than 10!&lt;/p&gt;
#     } %else {
#         &lt;p&gt;The value &lt;$value&gt; is less than 10!&lt;/p&gt;
#     }%&gt;
# </pre></blockquote>
#
# The example above includes different text in the middle of the
# dynamically generated HTML depending on whether TCL variable $value
# is less than 10.
#
# Notice that the keyword <b>%else</b> (with a percent sign in front)
# is used in place of the <b>else</b> keyword (without a percent sign.)
# You can also use <b>%elseif</b>, <b>%elseifdef</b>, and/or
# <b>%elseifndef</b>.
#
# The <b>%ifdef</b> and <b>%ifndef</b> variants of this command use as
# their conditional a list of indices to the <b>Q</b> array variable.
# <b>%ifdef</b> is successful if all indices named exist in <b>Q</b>
# and <b>%ifndef</b> is successful if none of the named indices are
# present in <b>Q</b>.  Thus, for example, the phrase
#
# <blockquote><pre>
# &lt;%%ifdef {a b} {Hello, World!}%&gt;
# </pre></blockquote>
#
# is equivalent to
#
# <blockquote><pre>
# &lt;%%if {[info exists Q(a)] && [info exists Q(b)]} {Hello, World!}%&gt;
# </pre></blockquote>
#
# The <b>%elseifdef</b> and <b>%elseifndef</b> keywords are similar
# extensions of <b>%elseif</b>.
#
proc %if     {args} {return [::tws::_if_processor [concat %elseif     $args]]}
proc %ifdef  {args} {return [::tws::_if_processor [concat %elseifdef  $args]]}
proc %ifndef {args} {return [::tws::_if_processor [concat %elseifndef $args]]}

# tclcmd: ifdef LIST
# title: Check to see if query parameters exist
#
# This command 
# returns non-zero (TRUE) if all of the arguments are query parameters.
# The test looks at the <b>Q</b> array variable in the calling procedure.
#
# For example, the command
#
# <blockquote><pre>
# if {[ifdef {a b c}]} break
# </pre></blockquote>
#
# is equivalent to
#
# <blockquote><pre>
# if {[info exists Q(a)] && [info exists Q(b)] && [info exists Q(c)]} break
# </pre></blockquote>
#
# This command returns TRUE if all named query parameters exists.
# Compare this to the <b>ifanydef</b> command which returns TRUE if one
# or more of the named query parameters exist. 
#
proc ifdef {args} {
  upvar 1 Q Q
  foreach q $args  {
    if {![info exists Q($q)]} {return 0}
  }
  return 1
}

# tclcmd: ifanydef LIST
# title: Check to see if any named query parameter exists
#
# This command 
# returns non-zero (TRUE) if any one of the arguments are query parameters.
# The test looks at the <b>Q</b> array variable in the calling procedure.
#
# For example, the command
#
# <blockquote><pre>
# if {[ifanydef {a b c}]} break
# </pre></blockquote>
#
# is equivalent to
#
# <blockquote><pre>
# if {[info exists Q(a)] || [info exists Q(b)] || [info exists Q(c)]} break
# </pre></blockquote>
#
# This command returns TRUE if one or more query parameters exists.
# Compare this to the <b>ifdef</b> command which returns TRUE if all
# named query parameters exist. 
#
proc ifanydef {args} {
  upvar 1 Q Q
  foreach q $args  {
    if {[info exists Q($q)]} {return 1}
  }
  return 0
}

# tclcmd: debug TEXT
# title: Add debugging output to the beginning of the returned HTML
#
# This command is useful for debugging embedded scripts.
# The argument is HTML text which will be prepended to the HTML that
# the web server generates and returns.  
#
# A common use for this command is to put short message in 
# embedded scripts to trace the generation of dynamic content.  Or
# it can be used to dump the values of variables.  For example, to
# view all the content of the <b>D</b> array variable, add this code
# anywhere in a dynamically generated document:
#
# <blockquote><pre>
# &lt;% ::tws::debug [::tws::parray D] %&gt;
# </pre></blockquote>
#
proc debug {text} {
  upvar 1 D D
  append D(debug) $text\n
}

# tclcmd: vardump VAR ...
# title: Prepend the values of variables to the returned HTML
#
# This command is useful for debugging embedded scripts.
# The arguments are the names of variable whose values are to be
# displayed.  The named variables can be scalars or arrays or a 
# mixture of the two.
#
# For example, to see the values of all query parameters and cookies,
# one could insert the following code:
#
# <blockquote><pre>
# &lt;% ::tws::vardump Q %&gt;
# </pre></blockquote>
#
proc vardump {args} {
  upvar 1 D D
  foreach vname $args {
    if {![catch {uplevel 1 [list set $vname]} val]} {
      append D(debug) [htmlize "$vname = $val "]<br>\n
    } else {
      foreach a [lsort -dictionary [uplevel 1 [list array names $vname]]] {
        if {$a=="debug" && $vname=="D"} continue
        set val [uplevel 1 [list set ${vname}($a)]]
        append D(debug) [htmlize "${vname}($a) = $val "]<br>\n
      }
    }
  }
}

# tclcmd: parray VARIABLE
# title: List the content of an array variable
#
# This command is similar to the <b>parray</b> command that is built into
# TCL.  The difference is that built-in version of <b>parray</b> writes
# the results to standard output whereas this version returns the
# result. 
#
# This command is intended to be used for debugging, such as in the
# following:
#
# <blockquote><pre>
# &lt;% ::tws::debug [::tws::parray D] %&gt;
# </pre></blockquote>
#
proc parray {vname {pattern {}}} {
  upvar 1 $vname V
  set x {}
  foreach e [lsort -dict [uplevel 1 "array names [list $vname] $pattern"]] {
    append x "[htmlize "${vname}($e) = $V($e)"] <br>\n"
  }
  return $x
}

# Export the those procedures that are frequently useful in the Tcl code
# of dynamic content.
#
namespace export %* include select require redirect
namespace export ifdef ifanydef setcookie
namespace export htmlize urlize httpize render hsubst

# End of the "namespace eval tws"
}
